unit CEFControl;

{$mode objfpc}
{$h+}

interface

uses
  Classes, uCEFTypes, uCEFApplication, uCEFWorkScheduler, uCEFInterfaces, uCEFResourceHandler;

Type

  { TLocalResourceHandler }

  TLocalResourceHandler = Class(TCefResourceHandlerOwn)
  private
    FFileName : String;
    FStream : TStream;
  Public
    function  ProcessRequest(const CEFRequest: ICefRequest; const callback: ICefCallback): Boolean; override;
    procedure GetResponseHeaders(const CEFResponse: ICefResponse; out responseLength: Int64; out redirectUrl: ustring); override;
    function  ReadResponse(const dataOut: Pointer; bytesToRead: Integer; var bytesRead: Integer; const callback: ICefCallback): Boolean; override;
    Destructor destroy; override;
  end;



Procedure InitCEF;
Procedure StopCEF;

implementation
//uCEFSchemeRegistrar, uCEFConstants,uCEFSchemeHandlerFactory,
//uCEFInterfaces, uCEFTypes, uCEFResourceHandler, uCEFMiscFunctions;


Uses SysUtils, uCEFConstants, uCEFSchemeRegistrar, uCEFMiscFunctions, uCEFLazarusCocoa; // required for Cocoa, AddCrDelegate

procedure GlobalCEFApp_OnScheduleMessagePumpWork(const aDelayMS : int64);
begin
  if (GlobalCEFWorkScheduler <> nil) then GlobalCEFWorkScheduler.ScheduleMessagePumpWork(aDelayMS);
end;


procedure CEFRegisterCustomSchemes(const registrar: TCefSchemeRegistrarRef);

Var
  aOptions : TCefSchemeOptions;

begin
  aOptions:=CEF_SCHEME_OPTION_STANDARD or CEF_SCHEME_OPTION_LOCAL;
  registrar.AddCustomScheme('local', aOptions);
end;


procedure CreateGlobalCEFApp;
begin
  if GlobalCEFApp <> nil then
    exit;
  GlobalCEFWorkScheduler := TCEFWorkScheduler.Create(nil);
  GlobalCEFApp:= TCefApplication.Create;
  GlobalCEFApp.ExternalMessagePump:= True;
  GlobalCEFApp.MultiThreadedMessageLoop:= False;
  GlobalCEFApp.OnRegCustomSchemes:=@CEFRegisterCustomSchemes;
  GlobalCEFApp.OnScheduleMessagePumpWork:= @GlobalCEFApp_OnScheduleMessagePumpWork;
  {$IFDEF LINUX}
  GlobalCEFApp.DisableZygote := True;
  {$ENDIF}
end;


Procedure InitCEF;

Var
  Dir : String;

begin
  if GlobalCEFApp <> nil then
    exit;
  {$IFDEF DARWIN}  // $IFDEF MACOSX
  AddCrDelegate;
  {$ENDIF}
  CreateGlobalCEFApp;
  {$IFDEF WINDOWS}
  Dir:='c:\CEF\Current\Dist':
  {$ELSE}
  Dir:='/opt/CEF/current/dist';
  {$ENDIF}
  GlobalCEFApp.FrameworkDirPath:=Dir;
  GlobalCEFApp.ResourcesDirPath:=Dir;
  GlobalCEFApp.LocalesDirPath:=Dir+PathDelim+'locales';
  if not GlobalCEFApp.StartMainProcess then
    begin
    StopCEF;
    halt(0); // exit the subprocess
    end;
end;

Procedure StopCEF;

begin
  if GlobalCEFWorkScheduler <> nil then
    GlobalCEFWorkScheduler.StopScheduler;
  DestroyGlobalCEFApp;
  DestroyGlobalCEFWorkScheduler;
end;

{ TLocalResourceHandler }

function TLocalResourceHandler.ProcessRequest(const CEFRequest: ICefRequest;
  const callback: ICefCallback): Boolean;

Var
  S : String;
  P : Integer;

begin
  S:=CEFRequest.GetUrl;
  P:=Pos('://',S);
  if P>0 then
    Delete(S,1,P+2);
  FFileName:=SysUtils.SetDirSeparators(ExtractFilePath(ParamStr(0))+S);
  Result:=FileExists(FFileName);
  if Result then
    FStream:=TFileStream.Create(FFileName,fmOpenRead or fmShareDenyWrite);
  if assigned(CallBack) then
    Callback.Cont;
end;

procedure TLocalResourceHandler.GetResponseHeaders(
  const CEFResponse: ICefResponse; out responseLength: Int64; out
  redirectUrl: ustring);

Var
  Ext : String;

begin
  Ext:=ExtractFileExt(FFileName);
  Ext:=Copy(Ext,2,Length(Ext)-1);
  if Assigned(FStream) then
    begin
    CefResponse.Status:=200;
    CefResponse.StatusText:='OK';
    responseLength:=FStream.Size;
    end
  else
    begin
    CefResponse.Status:=400;
    CefResponse.StatusText:='NOT FOUND';
    responseLength:=0;
    end;
  CefResponse.MimeType:=CefGetMimeType(Ext);
  // No other headers.
end;

function TLocalResourceHandler.ReadResponse(const dataOut: Pointer;
  bytesToRead: Integer; var bytesRead: Integer; const callback: ICefCallback
  ): Boolean;


Var
  aAvail : Integer;

begin
  Result:=Assigned(FStream) and Assigned(DataOut);
  if Result then
    begin
    aAvail:=FStream.Size-FStream.Position;
    If aAvail>BytesToRead then
      aAvail:=BytesToRead;
    BytesRead:=FStream.Read(DataOut^,aAvail);
    Result:=aAvail>0;
    end;
end;

destructor TLocalResourceHandler.destroy;
begin
  FreeAndNil(FStream);
  inherited destroy;
end;

initialization
{$IFDEF LINUX}
  InitCEF;
{$ENDIF}
end.

