unit CEFControl;

{$mode objfpc}
{$h+}

interface

uses
  uCEFApplication, uCEFWorkScheduler;


Procedure InitCEF;
Procedure StopCEF;

implementation

Uses SysUtils, uCEFLazarusCocoa; // required for Cocoa, AddCrDelegate

procedure GlobalCEFApp_OnScheduleMessagePumpWork(const aDelayMS : int64);
begin
  if (GlobalCEFWorkScheduler <> nil) then GlobalCEFWorkScheduler.ScheduleMessagePumpWork(aDelayMS);
end;

procedure CreateGlobalCEFApp;
begin
  if GlobalCEFApp <> nil then
    exit;
  GlobalCEFWorkScheduler := TCEFWorkScheduler.Create(nil);
  GlobalCEFApp:= TCefApplication.Create;
  GlobalCEFApp.ExternalMessagePump:= True;
  GlobalCEFApp.MultiThreadedMessageLoop:= False;
  GlobalCEFApp.OnScheduleMessagePumpWork:= @GlobalCEFApp_OnScheduleMessagePumpWork;
  {$IFDEF LINUX}
  GlobalCEFApp.DisableZygote := True;
  {$ENDIF}
end;


Procedure InitCEF;

Var
  Dir : String;

begin
  if GlobalCEFApp <> nil then
    exit;
  {$IFDEF DARWIN}  // $IFDEF MACOSX
  AddCrDelegate;
  {$ENDIF}
  CreateGlobalCEFApp;
  {$IFDEF WINDOWS}
  Dir:='c:\CEF\Current\Dist':
  {$ELSE}
  Dir:='/opt/CEF/current/dist';
  {$ENDIF}
  GlobalCEFApp.FrameworkDirPath:=Dir;
  GlobalCEFApp.ResourcesDirPath:=Dir;
  GlobalCEFApp.LocalesDirPath:=Dir+PathDelim+'locales';
  if not GlobalCEFApp.StartMainProcess then
    begin
    StopCEF;
    halt(0); // exit the subprocess
    end;
end;

Procedure StopCEF;

begin
  if GlobalCEFWorkScheduler <> nil then
    GlobalCEFWorkScheduler.StopScheduler;
  DestroyGlobalCEFApp;
  DestroyGlobalCEFWorkScheduler;
end;

initialization
{$IFDEF LINUX}
  InitCEF;
{$ENDIF}
end.

