unit frmMain;

{$mode objfpc}{$H+}

interface

uses
  SysUtils, Messages, Forms, Controls, Dialogs, ExtCtrls, StdCtrls, LMessages,
  Buttons, Menus, uCEFTypes, uCEFInterfaces, uCEFWorkScheduler,
  uCEFBrowserWindow, uCEFChromium, Classes, uCEFChromiumEvents;

type

  { TMainForm }

  TMainForm = class(TForm)
    btnDevTools: TBitBtn;
    btnForward: TBitBtn;
    btnBack: TBitBtn;
    edtURL: TEdit;
    btnGO: TButton;
    miHideDevTools: TMenuItem;
    miSHowDevTools: TMenuItem;
    pnlDevtoolsInner: TPanel;
    pnlDevtools: TPanel;
    pnlURL: TPanel;
    bwBrowser: TBrowserWindow;
    pmMenu: TPopupMenu;
    splDevtools: TSplitter;
    procedure btnBackClick(Sender: TObject);
    procedure btnDevToolsClick(Sender: TObject);
    procedure btnForwardClick(Sender: TObject);
    procedure ChromiumTitleChange(Sender: TObject; const browser: ICefBrowser;
      const title: ustring);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure btnGOClick(Sender: TObject);
    procedure bwBrowserBrowserClosed(Sender: TObject);
    procedure bwBrowserBrowserCreated(Sender: TObject);
    procedure miHideDevToolsClick(Sender: TObject);
    procedure miSHowDevToolsClick(Sender: TObject);
    procedure pmMenuPopup(Sender: TObject);
  private
    FNewTitle: string;
    procedure DoDevtoolsDetached(Sender: TObject; const browser: ICefBrowser);
    procedure DoSetHTMLTitle(Data: PtrInt);
  protected
    {$IFDEF WINDOWS}
    procedure WMEnterMenuLoop(var aMessage: TMessage); message WM_ENTERMENULOOP;
    procedure WMExitMenuLoop(var aMessage: TMessage); message WM_EXITMENULOOP;
    {$ENDIF}
  end;

var
  MainForm: TMainForm;

implementation

{$R *.lfm}

uses
  CEFControl;

{ TMainForm }

procedure TMainForm.btnGOClick(Sender: TObject);
begin
  bwBrowser.LoadURL(UTF8Decode(edtURL.Text));
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  splDevtools.Visible:=False;
  pnlDevtools.Visible:=False;
  bwBrowser.LoadURL(UTF8Decode(edtURL.Text));
end;

procedure TMainForm.bwBrowserBrowserCreated(Sender: TObject);
begin
  Caption := 'Browser ready';
end;

procedure TMainForm.miHideDevToolsClick(Sender: TObject);
begin
  bwBrowser.Chromium.CloseDevTools(pnlDevToolsInner);
  splDevtools.Visible:=False;
  pnlDevtools.Visible:=False;
end;

procedure TMainForm.miSHowDevToolsClick(Sender: TObject);
begin
  splDevtools.Visible:=True;
  pnlDevtools.Visible:=True;
  pnlDevtools.Height:=ClientHeight div 3;
  bwBrowser.Chromium.ShowDevTools(Point(0,0),pnlDevToolsInner);
  bwBrowser.Chromium.OnDevToolsAgentDetached:=@DoDevtoolsDetached;
end;

procedure TMainForm.pmMenuPopup(Sender: TObject);
begin
  miSHowDevTools.Enabled:=not pnlDevtools.Visible;
  miHideDevTools.Enabled:=pnlDevtools.Visible;
end;

procedure TMainForm.DoDevtoolsDetached(Sender: TObject;
  const browser: ICefBrowser);
begin
  pnlDevtools.Visible:=False;
  splDevtools.Visible:=False;
end;

procedure TMainForm.DoSetHTMLTitle(Data: PtrInt);
begin
  Caption:=FNewTitle;
end;

{$IFDEF WINDOWS}
procedure TMainForm.WMEnterMenuLoop(var aMessage: TMessage);
begin
  inherited;
  if (aMessage.wParam = 0) and (GlobalCEFApp <> nil) then
    GlobalCEFApp.OsmodalLoop := True;
end;

procedure TMainForm.WMExitMenuLoop(var aMessage: TMessage);
begin
  inherited;
  if (aMessage.wParam = 0) and (GlobalCEFApp <> nil) then
    GlobalCEFApp.OsmodalLoop := False;
end;
{$ENDIF}

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  With bwBrowser do
    begin
    CloseBrowser(True);
    CanClose:=IsClosed;
    end;
end;

procedure TMainForm.btnBackClick(Sender: TObject);
begin
  if bwbrowser.Chromium.CanGoBack then
    bwbrowser.Chromium.GoBack;
end;

procedure TMainForm.btnDevToolsClick(Sender: TObject);

Var
  P : TPoint;

begin
  With bwBrowser.BoundsRect do
    P:=Point(bwBrowser.BoundsRect.Right,bwBrowser.BoundsRect.Top);
  P:=ControlToScreen(P);
  pmMenu.PopUp(P.X-200,P.Y);
end;

procedure TMainForm.btnForwardClick(Sender: TObject);
begin
  if bwbrowser.Chromium.CanGoForward then
    bwbrowser.Chromium.GoForward;
end;

procedure TMainForm.ChromiumTitleChange(Sender: TObject;
  const browser: ICefBrowser; const title: ustring);
begin
  if (length(title) > 0) then
    FNewTitle := UTF8enCode(title)
   else
    FNewTitle:=UTF8enCode(bwBrowser.Chromium.DocumentURL);
  Application.QueueAsyncCall(@DoSetHTMLTitle,0);
end;

procedure TMainForm.bwBrowserBrowserClosed(Sender: TObject);
begin
  Close;
end;


initialization
  InitCEF;

finalization
  StopCEF;
end.

